# pylint: disable=W0611
# flake8: noqa
from . import diagnostic
from .diagnostic import (
    acorr_ljungbox, acorr_breusch_godfrey,
    CompareCox, compare_cox, CompareJ, compare_j,
    HetGoldfeldQuandt, het_goldfeldquandt,
    het_breuschpagan, het_white, het_arch,
    linear_harvey_collier, linear_rainbow, linear_lm,
    breaks_cusumolsresid, breaks_hansen, recursive_olsresiduals,
    unitroot_adf,
    normal_ad, lilliefors
)

from ._knockoff import RegressionFDR
from . import multicomp
from .multitest import (multipletests, fdrcorrection, fdrcorrection_twostage,
                        local_fdr, NullDistribution)
from .multicomp import tukeyhsd
from . import gof
from .gof import (powerdiscrepancy, gof_chisquare_discrete,
                  chisquare_effectsize)
from . import stattools
from .stattools import durbin_watson, omni_normtest, jarque_bera

from . import sandwich_covariance
from .sandwich_covariance import (
    cov_cluster, cov_cluster_2groups, cov_nw_panel,
    cov_hac, cov_white_simple,
    cov_hc0, cov_hc1, cov_hc2, cov_hc3,
    se_cov
)

from .weightstats import (DescrStatsW, CompareMeans, ttest_ind, ttost_ind,
                          ttost_paired, ztest, ztost, zconfint)

from .proportion import (
    binom_test_reject_interval, binom_test,
    binom_tost, binom_tost_reject_interval,
    power_binom_tost, power_ztost_prop,
    proportion_confint, proportion_effectsize,
    proportions_chisquare, proportions_chisquare_allpairs,
    proportions_chisquare_pairscontrol, proportions_ztest,
    proportions_ztost, multinomial_proportions_confint)

from .power import (TTestPower, TTestIndPower, GofChisquarePower,
                    NormalIndPower, FTestAnovaPower, FTestPower,
                    tt_solve_power, tt_ind_solve_power, zt_ind_solve_power)

from .descriptivestats import Describe

from .anova import anova_lm

from . import moment_helpers
from .correlation_tools import (
    corr_clipped, corr_nearest,
    corr_nearest_factor, corr_thresholded, cov_nearest,
    cov_nearest_factor_homog, FactoredPSDMatrix)

from statsmodels.sandbox.stats.runs import Runs, runstest_1samp, runstest_2samp

from statsmodels.stats.contingency_tables import (mcnemar, cochrans_q,
                                                  SquareTable,
                                                  Table2x2,
                                                  Table,
                                                  StratifiedTable)
from .mediation import Mediation



# -*- coding: utf-8 -*-

"""Multivariate analysis of variance

author: Yichuan Liu
"""
from __future__ import division

import numpy as np

from statsmodels.base.model import Model
from .multivariate_ols import MultivariateTestResults
from .multivariate_ols import _multivariate_ols_fit
from .multivariate_ols import _multivariate_ols_test, _hypotheses_doc

__docformat__ = 'restructuredtext en'


class MANOVA(Model):
    """
    Multivariate analysis of variance

    The implementation of MANOVA is based on multivariate regression and does
    not assume that the explanatory variables are categorical. Any type of
    variables as in regression is allowed.

    Parameters
    ----------
    endog : array_like
        Dependent variables. A nobs x k_endog array where nobs is
        the number of observations and k_endog is the number of dependent
        variables.
    exog : array_like
        Independent variables. A nobs x k_exog array where nobs is the
        number of observations and k_exog is the number of independent
        variables. An intercept is not included by default and should be added
        by the user. Models specified using a formula include an intercept by
        default.

    Attributes
    ----------
    endog : array
        See Parameters.
    exog : array
        See Parameters.

    Notes
    -----
    MANOVA is used though the `mv_test` function, and `fit` is not used.

    The ``from_formula`` interface is the recommended method to specify
    a model and simplifies testing without needing to manually configure
    the contrast matrices.

    References
    ----------
    .. [*] ftp://public.dhe.ibm.com/software/analytics/spss/documentation/statistics/20.0/en/client/Manuals/IBM_SPSS_Statistics_Algorithms.pdf
    """
    def __init__(self, endog, exog, missing='none', hasconst=None, **kwargs):
        if len(endog.shape) == 1 or endog.shape[1] == 1:
            raise ValueError('There must be more than one dependent variable'
                             ' to fit MANOVA!')
        super(MANOVA, self).__init__(endog, exog, missing=missing,
                                     hasconst=hasconst, **kwargs)
        self._fittedmod = _multivariate_ols_fit(self.endog, self.exog)

    def fit(self):
        raise NotImplementedError('fit is not needed to use MANOVA. Call'
                                  'mv_test directly on a MANOVA instance.')

    def mv_test(self, hypotheses=None):
        if hypotheses is None:
            if (hasattr(self, 'data') and self.data is not None and
                        hasattr(self.data, 'design_info')):
                terms = self.data.design_info.term_name_slices
                hypotheses = []
                for key in terms:
                    L_contrast = np.eye(self.exog.shape[1])[terms[key], :]
                    hypotheses.append([key, L_contrast, None])
            else:
                hypotheses = []
                for i in range(self.exog.shape[1]):
                    name = 'x%d' % (i)
                    L = np.zeros([1, self.exog.shape[1]])
                    L[0, i] = 1
                    hypotheses.append([name, L, None])

        results = _multivariate_ols_test(hypotheses, self._fittedmod,
                                         self.exog_names, self.endog_names)

        return MultivariateTestResults(results, self.endog_names,
                                       self.exog_names)
    mv_test.__doc__ = (
"""
Linear hypotheses testing

Parameters
----------
""" + _hypotheses_doc + """

Returns
-------
results: MultivariateTestResults

Notes
-----
Testing the linear hypotheses

    L * params * M = 0

where `params` is the regression coefficient matrix for the
linear model y = x * params

If the model is not specified using the formula interfact, then the hypotheses
test each included exogenous variable, one at a time. In most applications
with categorical variables, the ``from_formula`` interface should be preferred
when specifying a model since it provides knowledge about the model when
specifying the hypotheses.
""")



from numpy.testing import assert_
import pytest

from statsmodels.tsa.x13 import _find_x12, x13_arima_select_order

x13path = _find_x12()

pytestmark = pytest.mark.skipif(x13path is False, reason='X13/X12 not available')


class TestX13(object):
    @classmethod
    def setup_class(cls):
        import pandas as pd
        from statsmodels.datasets import macrodata, co2
        dta = macrodata.load_pandas().data
        index = pd.period_range(start='1959Q1', end='2009Q3', freq='Q')
        dta.index = index
        cls.quarterly_data = dta.dropna()

        dta = co2.load_pandas().data
        dta['co2'] = dta.co2.interpolate()
        cls.monthly_data = dta.resample('M')
        # change in pandas 0.18 resample is deferred object
        if not isinstance(cls.monthly_data, (pd.DataFrame, pd.Series)):
            cls.monthly_data = cls.monthly_data.mean()

        cls.monthly_start_data = dta.resample('MS')
        if not isinstance(cls.monthly_start_data, (pd.DataFrame, pd.Series)):
            cls.monthly_start_data = cls.monthly_start_data.mean()

    def test_x13_arima_select_order(self):
        res = x13_arima_select_order(self.monthly_data)
        assert_(isinstance(res.order, tuple))
        assert_(isinstance(res.sorder, tuple))
        res = x13_arima_select_order(self.monthly_start_data)
        assert_(isinstance(res.order, tuple))
        assert_(isinstance(res.sorder, tuple))
        res = x13_arima_select_order(self.monthly_data.co2)
        assert_(isinstance(res.order, tuple))
        assert_(isinstance(res.sorder, tuple))
        res = x13_arima_select_order(self.monthly_start_data.co2)
        assert_(isinstance(res.order, tuple))
        assert_(isinstance(res.sorder, tuple))

        res = x13_arima_select_order(self.quarterly_data[['realgdp']])
        assert_(isinstance(res.order, tuple))
        assert_(isinstance(res.sorder, tuple))
        res = x13_arima_select_order(self.quarterly_data.realgdp)
        assert_(isinstance(res.order, tuple))
        assert_(isinstance(res.sorder, tuple))



__all__ = ["AR", "ARMA", "ARIMA",
           "var", "VAR", "VECM", "SVAR", "DynamicVAR",
           "filters",
           "innovations",
           "tsatools",
           "add_trend", "detrend", "lagmat", "lagmat2ds", "add_lag",
           "interp",
           "stattools",
           "acovf", "acf", "pacf", "pacf_yw", "pacf_ols", "ccovf", "ccf",
           "periodogram", "q_stat", "coint", "arma_order_select_ic",
           "adfuller", "kpss", "bds",
           "datetools",
           "seasonal_decompose",
           "graphics",
           "x13_arima_select_order", "x13_arima_analysis",
           "statespace",
           "SARIMAX", "UnobservedComponents", "VARMAX", "DynamicFactor",
           "MarkovRegression", "MarkovAutoregression",
           "ExponentialSmoothing", "SimpleExpSmoothing", "Holt",
           "arma_generate_sample", "ArmaProcess"]

from .ar_model import AR
from .arima_model import ARMA, ARIMA
from . import vector_ar as var
from .arima_process import arma_generate_sample, ArmaProcess
from .vector_ar.var_model import VAR
from .vector_ar.vecm import VECM
from .vector_ar.svar_model import SVAR
from .vector_ar.dynamic import DynamicVAR
from .filters import api as filters
from . import tsatools
from .tsatools import (add_trend, detrend, lagmat, lagmat2ds, add_lag)
from . import interp
from . import stattools
from .stattools import (
    acovf, acf, pacf, pacf_yw, pacf_ols, ccovf, ccf,
    periodogram, q_stat, coint, arma_order_select_ic,
    adfuller, kpss, bds)
from .base import datetools
from .seasonal import seasonal_decompose
from ..graphics import tsaplots as graphics
from .x13 import x13_arima_select_order
from .x13 import x13_arima_analysis
from .statespace import api as statespace
from .statespace.sarimax import SARIMAX
from .statespace.structural import UnobservedComponents
from .statespace.varmax import VARMAX
from .statespace.dynamic_factor import DynamicFactor
from .regime_switching.markov_regression import MarkovRegression
from .regime_switching.markov_autoregression import MarkovAutoregression
from .holtwinters import ExponentialSmoothing, SimpleExpSmoothing, Holt
from .innovations import api as innovations



from __future__ import print_function
from statsmodels.tools.sm_exceptions import CacheWriteWarning
import warnings

__all__ = ['cache_readonly', 'cache_writable']


class CachedAttribute(object):

    def __init__(self, func, cachename=None):
        self.fget = func
        self.name = func.__name__
        self.cachename = cachename or '_cache'

    def __get__(self, obj, type=None):
        if obj is None:
            return self.fget
        # Get the cache or set a default one if needed
        _cachename = self.cachename
        _cache = getattr(obj, _cachename, None)
        if _cache is None:
            setattr(obj, _cachename, {})
            _cache = getattr(obj, _cachename)
        # Get the name of the attribute to set and cache
        name = self.name
        _cachedval = _cache.get(name, None)
        if _cachedval is None:
            _cachedval = self.fget(obj)
            _cache[name] = _cachedval

        return _cachedval

    def __set__(self, obj, value):
        errmsg = "The attribute '%s' cannot be overwritten" % self.name
        warnings.warn(errmsg, CacheWriteWarning)


class CachedWritableAttribute(CachedAttribute):
    def __set__(self, obj, value):
        _cache = getattr(obj, self.cachename)
        name = self.name
        _cache[name] = value


class _cache_readonly(property):
    """
    Decorator for CachedAttribute
    """

    def __init__(self, cachename=None):
        self.func = None
        self.cachename = cachename

    def __call__(self, func):
        return CachedAttribute(func,
                               cachename=self.cachename)


cache_readonly = _cache_readonly()


class cache_writable(_cache_readonly):
    """
    Decorator for CachedWritableAttribute
    """
    def __call__(self, func):
        return CachedWritableAttribute(func,
                                       cachename=self.cachename)


def nottest(fn):
    fn.__test__ = False
    return fn



import numpy as np
from statsmodels.base import model
import statsmodels.base.wrapper as wrap


class _DimReductionRegression(model.Model):
    """
    A base class for dimension reduction regression methods.
    """

    def __init__(self, endog, exog, **kwargs):
        super(_DimReductionRegression, self).__init__(endog, exog, **kwargs)

    def _prep(self, n_slice):

        # Sort the data by endog
        ii = np.argsort(self.endog)
        x = self.exog[ii, :]

        # Whiten the data
        x -= x.mean(0)
        covx = np.cov(x.T)
        covxr = np.linalg.cholesky(covx)
        x = np.linalg.solve(covxr, x.T).T
        self.wexog = x
        self._covxr = covxr

        # Split the data into slices
        self._split_wexog = np.array_split(x, n_slice)


class SlicedInverseReg(_DimReductionRegression):
    """
    Sliced Inverse Regression (SIR)

    Parameters
    ----------
    endog : array-like (1d)
        The dependent variable
    exog : array-like (2d)
        The covariates

    References
    ----------
    KC Li (1991).  Sliced inverse regression for dimension reduction.
    JASA 86, 316-342.
    """

    def fit(self, **kwargs):
        """
        Estimate the EDR space.

        Parameters
        ----------
        slice_n : int, optional
            Number of observations per slice
        """

        # Sample size per slice
        slice_n = kwargs.get("slice_n", 20)

        # Number of slices
        n_slice = self.exog.shape[0] // slice_n

        self._prep(n_slice)

        mn = [z.mean(0) for z in self._split_wexog]
        n = [z.shape[0] for z in self._split_wexog]
        mn = np.asarray(mn)
        n = np.asarray(n)
        mnc = np.cov(mn.T, fweights=n)

        a, b = np.linalg.eigh(mnc)
        jj = np.argsort(-a)
        a = a[jj]
        b = b[:, jj]
        params = np.linalg.solve(self._covxr.T, b)

        results = DimReductionResults(self, params, eigs=a)
        return DimReductionResultsWrapper(results)


class PrincipalHessianDirections(_DimReductionRegression):
    """
    Principal Hessian Directions

    Parameters
    ----------
    endog : array-like (1d)
        The dependent variable
    exog : array-like (2d)
        The covariates

    References
    ----------
    KC Li (1992).  On Principal Hessian Directions for Data
    Visualization and Dimension Reduction: Another application
    of Stein's lemma. JASA 87:420.
    """

    def fit(self, **kwargs):
        """
        Estimate the EDR space using PHD.

        Parameters
        ----------
        resid : bool, optional
            If True, use least squares regression to remove the
            linear relationship between each covariate and the
            response, before conducting PHD.
        """

        resid = kwargs.get("resid", False)

        y = self.endog - self.endog.mean()
        x = self.exog - self.exog.mean(0)

        if resid:
            from statsmodels.regression.linear_model import OLS
            r = OLS(y, x).fit()
            y = r.resid

        cm = np.einsum('i,ij,ik->jk', y, x, x)
        cm /= len(y)

        cx = np.cov(x.T)
        cb = np.linalg.solve(cx, cm)

        a, b = np.linalg.eig(cb)
        jj = np.argsort(-np.abs(a))
        a = a[jj]
        params = b[:, jj]

        results = DimReductionResults(self, params, eigs=a)
        return DimReductionResultsWrapper(results)


class SlicedAverageVarianceEstimation(_DimReductionRegression):
    """
    Sliced Average Variance Estimation (SAVE)

    Parameters
    ----------
    endog : array-like (1d)
        The dependent variable
    exog : array-like (2d)
        The covariates
    bc : bool, optional
        If True, use the bias-correctedCSAVE method of Li and Zhu.

    References
    ----------
    RD Cook.  SAVE: A method for dimension reduction and graphics
    in regression.
    http://www.stat.umn.edu/RegGraph/RecentDev/save.pdf

    Y Li, L-X Zhu (2007). Asymptotics for sliced average
    variance estimation.  The Annals of Statistics.
    https://arxiv.org/pdf/0708.0462.pdf
    """

    def __init__(self, endog, exog, **kwargs):
        super(SAVE, self).__init__(endog, exog, **kwargs)

        self.bc = False
        if "bc" in kwargs and kwargs["bc"] is True:
            self.bc = True

    def fit(self, **kwargs):
        """
        Estimate the EDR space.

        Parameters
        ----------
        slice_n : int
            Number of observations per slice
        """

        # Sample size per slice
        slice_n = kwargs.get("slice_n", 50)

        # Number of slices
        n_slice = self.exog.shape[0] // slice_n

        self._prep(n_slice)

        cv = [np.cov(z.T) for z in self._split_wexog]
        ns = [z.shape[0] for z in self._split_wexog]

        p = self.wexog.shape[1]

        if not self.bc:
            # Cook's original approach
            vm = 0
            for w, cvx in zip(ns, cv):
                icv = np.eye(p) - cvx
                vm += w * np.dot(icv, icv)
            vm /= len(cv)
        else:
            # The bias-corrected approach of Li and Zhu

            # \Lambda_n in Li, Zhu
            av = 0
            for c in cv:
                av += np.dot(c, c)
            av /= len(cv)

            # V_n in Li, Zhu
            vn = 0
            for x in self._split_wexog:
                r = x - x.mean(0)
                for i in range(r.shape[0]):
                    u = r[i, :]
                    m = np.outer(u, u)
                    vn += np.dot(m, m)
            vn /= self.exog.shape[0]

            c = np.mean(ns)
            k1 = c * (c - 1) / ((c - 1)**2 + 1)
            k2 = (c - 1) / ((c - 1)**2 + 1)
            av2 = k1 * av - k2 * vn

            vm = np.eye(p) - 2 * sum(cv) / len(cv) + av2

        a, b = np.linalg.eigh(vm)
        jj = np.argsort(-a)
        a = a[jj]
        b = b[:, jj]
        params = np.linalg.solve(self._covxr.T, b)

        results = DimReductionResults(self, params, eigs=a)
        return DimReductionResultsWrapper(results)


class DimReductionResults(model.Results):
    """
    Results class for a dimension reduction regression.
    """

    def __init__(self, model, params, eigs):
        super(DimReductionResults, self).__init__(
              model, params)
        self.eigs = eigs


class DimReductionResultsWrapper(wrap.ResultsWrapper):
    _attrs = {
        'params': 'columns',
    }
    _wrap_attrs = _attrs

wrap.populate_wrapper(DimReductionResultsWrapper,  # noqa:E305
                      DimReductionResults)

# aliases for expert users
SIR = SlicedInverseReg
PHD = PrincipalHessianDirections
SAVE = SlicedAverageVarianceEstimation



# -*- coding: utf-8 -*-

import pytest
from numpy.testing import assert_equal

from statsmodels.tools.decorators import (
    cache_readonly, CacheWriteWarning)


def test_cache_readonly():

    class Example(object):
        def __init__(self):
            self._cache = {}
            self.a = 0

        @cache_readonly
        def b(self):
            return 1

        @cache_readonly
        def e(self):
            return 4

        @cache_readonly
        def f(self):
            return self.e + 1

    ex = Example()

    # Try accessing/setting a readonly attribute
    assert_equal(ex.__dict__, dict(a=0, _cache={}))

    b = ex.b
    assert_equal(b, 1)
    assert_equal(ex.__dict__, dict(a=0, _cache=dict(b=1,)))
    # assert_equal(ex.__dict__, dict(a=0, b=1, _cache=dict(b=1)))

    with pytest.warns(CacheWriteWarning):
        ex.b = -1

    assert_equal(ex._cache, dict(b=1,))



from __future__ import absolute_import

from distutils.version import LooseVersion

import pandas


version = LooseVersion(pandas.__version__)
pandas_lte_0_19_2 = version <= LooseVersion('0.19.2')
pandas_gt_0_19_2 = version > LooseVersion('0.19.2')


try:
    from pandas.api.types import is_numeric_dtype  # noqa:F401
except ImportError:
    from pandas.core.common import is_numeric_dtype  # noqa:F401

if version >= '0.20':
    try:
        from pandas.tseries import offsets as frequencies
    except ImportError:
        from pandas.tseries import frequencies
    data_klasses = (pandas.Series, pandas.DataFrame, pandas.Panel)
else:
    try:
        import pandas.tseries.frequencies as frequencies
    except ImportError:
        from pandas.core import datetools as frequencies  # noqa

    data_klasses = (pandas.Series, pandas.DataFrame, pandas.Panel,
                    pandas.WidePanel)

try:
    import pandas.testing as testing
except ImportError:
    import pandas.util.testing as testing

assert_frame_equal = testing.assert_frame_equal
assert_index_equal = testing.assert_index_equal
assert_series_equal = testing.assert_series_equal



import subprocess

import pytest

from statsmodels.compat import PY3
from statsmodels.compat.platform import PLATFORM_WIN


@pytest.mark.skipif(PLATFORM_WIN, reason="subprocess does not play nicely")
def test_lazy_imports():
    # Check that when statsmodels.api is imported, matplotlib is _not_ imported
    cmd = ("import statsmodels.api as sm; "
           "import sys; "
           "mods = [x for x in sys.modules if 'matplotlib.pyplot' in x]; "
           "assert not mods, mods")

    # TODO: is there a cleaner way to do this import in an isolated environment
    pyexe = 'python3' if PY3 else 'python'
    p = subprocess.Popen(pyexe + ' -c "' + cmd + '"',
                         shell=True, close_fds=True)
    p.wait()
    rc = p.returncode
    assert rc == 0
